
/*

	Colors:         0 Black         1 Blue          2 Green
			3 Cyan          4 deep red      5 Magenta
			6 med Red       7 Lt. Gray      8 Dk. Gray
			9 Brite Blue    10 Brite green  11 Brite Cyan
			12 Brite red    13 Brite Magen  14 yellow
					15 White






	dxf2v -- strip 3d faces out of a .dfx file

	Copyright 1991 Stephen B. Coy
	 modified by   Tom Barber    april 3 1992  added path recognition
						   added layer extraction

	 modified by   Mike Daigle   july 3 1992   small fix to layer
							 extraction
						   added surface color
							 extraction
	 modified by   Adam Shiffman july 30 1992  added Vivid 2.0
							output

*/


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
	char *colors [16] = { "black","blue","green","cyan","brown",
			     "magenta","red","gray","dim_gray",
			     "light_blue","lime_green",
			     "salmon","orange_red","blue_violet",
			     "yellow","white", };


main(int argc, char *argv[])
{
	float   ambient_red,ambient_green,ambient_blue;
	char    str[80];
	int     key, npnts,layer,l,color,oldcolor;
	double  x1, x2, x3, x4,
		y1, y2, y3, y4,
		z1, z2, z3, z4;
	FILE    *dxf_in_file,*vivid_out_file;
		color = 15;


	if(argc < 3) {
		printf("\nDXF2V Version 2.1 - Adapted from Stephen Coy's DXF2V\n");
		printf("Converts 3DFACEs in DXF files to VIVID files of X,Y,Z points.\n");
		printf("Extracts by Layers (if desired) and Extracts colors (desired or not)\n");
		printf("Added code by Tom Barber 4/92  Mike Daigle 7/92\n");
		printf("Added Vivid 2.0 specific output 7/92 Adam Shiffman\n\n");
	printf("Usage: dXF2V file.dxf file.v [Layer Number] [Ambient Red] [Ambient Green]\n\t [Ambient Blue]\n\n");
		printf("Where:\n");
		printf("       file.dxf = name of dxf file with extension\n");
		printf("       file.v = name of VIVID file with extension\n");
		printf("       [Layer #]= Layer to be converted.  Must be specified,\n");
		printf("                  if Ambients are used. -1 is all layers.\n");
		printf("       [Ambient Red]   = Ambient Color Red Content default = 0\n");
		printf("       [Ambient Green] = Ambient Color Green Content deflt = 0\n");
		printf("       [Ambient Blue]  = Ambient Color Blue Content  deflt = 0\n");
		printf(" Typical: DXF2V art.dxf art.v -1 .1 .2 .3\n");
		printf("       convert art.dxf to art.v all layers red =.1 green = .2 blue = .3\n");
		exit(1);
	}

	//Open the files
	if (!(dxf_in_file = fopen(argv[1],"rb")))
		{ puts("Can not open DXF file!");
		  exit(1);
		}

	if (!(vivid_out_file = fopen(argv[2],"w")))
		{ puts("Can not open VIVID file for output!");
		  fclose(dxf_in_file);
		  exit(1);
		}

 ambient_red = 0.0;
 ambient_green = 0.0;
 ambient_blue = 0.0;
     if (argc > 4) ambient_red = atof(argv[4]);
     if (argc > 5) ambient_green = atof(argv[5]);
     if (argc > 6) ambient_blue = atof(argv[6]);

layer = -1;

if (argc > 3) layer = atoi(argv[3]);

	while(fgets(str, 80, dxf_in_file)) {
		//Wait until we find a 3DFACE...
		if(strncmp(str, "3DFACE", 6) != 0) {continue;}
		   fgets(str,80,dxf_in_file);
		   if (atoi(str) != 8) {printf("WARNING...LAYER...MISSED\n");}
		   fgets(str,80,dxf_in_file);

		   if (layer != atoi(str) && layer != -1)   {continue;}

		oldcolor = color;
		x1 = x2 = x3 = x4 = y1 = y2 = y3 = y4 = z1 = z2 = z3 = z4 = 0.0;
		do {
			fgets(str, 80, dxf_in_file);
			key = atoi(str);
			fgets(str, 80, dxf_in_file);
			switch(key) {
				case 62 :color = atoi(str); break;
				case 10 : x1 = atof(str); break;
				case 20 : y1 = atof(str); break;
				case 30 : z1 = atof(str); break;
				case 11 : x2 = atof(str); break;
				case 21 : y2 = atof(str); break;
				case 31 : z2 = atof(str); break;
				case 12 : x3 = atof(str); break;
				case 22 : y3 = atof(str); break;
				case 32 : z3 = atof(str); break;
				case 13 : x4 = atof(str); break;
				case 23 : y4 = atof(str); break;
				case 33 : z4 = atof(str); break;
				default : break;
			}
		} while(key != 33);

		/* check for non-unique points */

		npnts = 4;
		if(x1==x2 && y1==y2 && z1==z2) {
			x2 = x3; x3 = x4;
			y2 = y3; y3 = y4;
			z2 = z3; z3 = z4;
			npnts--;
		}
		if(npnts < 3) continue;
		if(x1==x2 && y1==y2 && z1==z2) {
			continue;
		}
		if(x1==x3 && y1==y3 && z1==z3) {
			x3 = x4;
			y3 = y4;
			z3 = z4;
			npnts--;
		}
		if(npnts < 3) continue;
		if(x2==x3 && y2==y3 && z2==z3) {
			x3 = x4;
			y3 = y4;
			z3 = z4;
			npnts--;
		}
		if(npnts < 3) continue;
		if(npnts==4 && x4==x3 && y4==y3 && z4==z3) {
			npnts--;
		}
		if(npnts < 3) continue;

		/* got the poly, output it */

	    /* write new surface if color changes */
	       while (color > 15)
		     {
		     color = color - 16;
		     }
	    if (oldcolor != color) {
		   fprintf(vivid_out_file,"surface { \n\tdiff  %s \n\tambient %f %f %f\n",colors[color],ambient_red,ambient_green,ambient_blue);
		   fprintf(vivid_out_file,"\tspec .2 .2 .2\n\tshine 100\n          }\n");
		   }
		if(npnts == 4) {
	    fprintf(vivid_out_file,"polygon { points 3\n");
	    fprintf(vivid_out_file,"\tvertex %.4f %.4f %.4f\n", x1, y1, z1);
	    fprintf(vivid_out_file,"\tvertex %.4f %.4f %.4f\n", x2, y2, z2);
	    fprintf(vivid_out_file,"\tvertex %.4f %.4f %.4f\n}\n", x3, y3, z3);
	    fprintf(vivid_out_file,"polygon { points 3\n");
	    fprintf(vivid_out_file,"\tvertex %.4f %.4f %.4f\n", x1, y1, z1);
	    fprintf(vivid_out_file,"\tvertex %.4f %.4f %.4f\n", x3, y3, z3);
	    fprintf(vivid_out_file,"\tvertex %.4f %.4f %.4f\n}\n", x4, y4, z4);
		} else {
	    fprintf(vivid_out_file,"polygon { points 3\n");
	    fprintf(vivid_out_file,"\tvertex %.4f %.4f %.4f\n", x1, y1, z1);
	    fprintf(vivid_out_file,"\tvertex %.4f %.4f %.4f\n", x2, y2, z2);
	    fprintf(vivid_out_file,"\tvertex %.4f %.4f %.4f\n}\n", x3, y3, z3);
	}   /* END WHILE */

   }

fclose(vivid_out_file);
fclose(dxf_in_file);
}       /* end of main() */
